/*
 * Copyright © 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Thomas Voß <thomas.voss@canonical.com>
 */
#ifndef CORE_DBUS_TRAITS_TIMEOUT_H_
#define CORE_DBUS_TRAITS_TIMEOUT_H_

#include <chrono>

namespace core
{
namespace dbus
{
namespace traits
{
template<typename UnderlyingType>
struct Timeout
{
    typedef std::chrono::milliseconds DurationType;

    static bool is_timeout_enabled(UnderlyingType* timeout);
    static DurationType get_timeout_interval(UnderlyingType* timeout);
    static void invoke_timeout_handler(UnderlyingType* timeout);
};
}
}
}

#endif // CORE_DBUS_TRAITS_WATCH_H_
